public class Procedure { 
    Program     prog;
    Procedure   top;
    String      name;
    String[]    vars;
    int         level;
    byte[]      code;
    Procedure[] subprocs;

    transient int nLabels;

    static final int MAX_VARIABLES = 0x100;
    static final int MAX_SUBPROCS  = 0x100;

    Procedure(Program prog, Procedure top) { 
        this.prog = prog;
        this.top = top;
    }        

    Procedure(Program prog, Procedure top, String name, int level) { 
        this(prog, top);
        this.name = name;
        this.level = level;
        code = new byte[1];
        code[0] = VirtualMachine.RETURN;
        vars = new String[0];
        subprocs = new Procedure[0];
    }
        
    boolean findProcedureCall(int level, int index) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.CALL) { 
                if (code[pc+1] == level && (code[pc+2] & 0xFF) == index) { 
                    return true;
                }
            }
        }
        for (int i = 0; i < subprocs.length; i++) { 
            if (subprocs[i].findProcedureCall(level, index)) { 
                return true;
            }
        }
        return false;
    }

    void adjustProcedureCall(int level, int index) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.CALL) { 
                if (code[pc+1] == level && (code[pc+2] & 0xFF) > index) { 
                    code[pc+2] -= 1;
                }
            }
        }
        for (int i = 0; i < subprocs.length; i++) { 
            subprocs[i].adjustProcedureCall(level, index);
        }
    }

    int findSubproc(String name) { 
        for (int i = 0; i < subprocs.length; i++) { 
            if (subprocs[i].name.equals(name)) { 
                return i;
            }
        }
        return -1;
    }
    
    int addSubproc(String name) { 
        if (findSubproc(name) >= 0) { 
            return -1;
        }
        int n = subprocs.length;
        Procedure[] newSubprocs = new Procedure[n+1];
        System.arraycopy(subprocs, 0, newSubprocs, 0, n);
        subprocs = newSubprocs;
        subprocs[n] = new Procedure(prog, this, name, level+1);
        return n;
    }

    boolean removeSubproc(int i) { 
        if (findProcedureCall(level, i)) { 
            return false;
        }
        int n = subprocs.length;
        Procedure[] newSubprocs = new Procedure[n-1];
        System.arraycopy(subprocs, 0, newSubprocs, 0, i);
        System.arraycopy(subprocs, i+1, newSubprocs, i, n-i-1);
        subprocs = newSubprocs;
        adjustProcedureCall(level, i);
        return true;
    }

    int findVariable(String name) { 
        for (int i = 0; i < vars.length; i++) { 
            if (vars[i].equals(name)) { 
                return i;
            }
        }
        return -1;
    }

    int addVariable(String name) { 
        if (findVariable(name) >= 0) { 
            return -1;
        }
        int n = vars.length;
        String[] newVars = new String[n+1];
        System.arraycopy(vars, 0, newVars, 0, n);
        vars = newVars;
        vars[n] = name;
        return n;
    }

    boolean findVariableAccess(int level, int index) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.LOAD || code[pc] == VirtualMachine.STORE) { 
                if (code[pc+1] == level && (code[pc+2] & 0xFF) == index) { 
                    return true;
                }
            }
        }
        for (int i = 0; i < subprocs.length; i++) { 
            if (subprocs[i].findVariableAccess(level, index)) { 
                return true;
            }
        }
        return false;
    }

    void adjustVariableAccess(int level, int index) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.LOAD || code[pc] == VirtualMachine.STORE) { 
                if (code[pc+1] == level && (code[pc+2] & 0xFF) > index) { 
                    code[pc+2] -= 1;
                }
            }
        }
        for (int i = 0; i < subprocs.length; i++) { 
            subprocs[i].adjustVariableAccess(level, index);
        }
    }

    boolean findParameterAccess(int index) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.INPUT) { 
                if ((code[pc+1] & 0xFF) == index) { 
                    return true;
                }
            }
        }
        for (int i = 0; i < subprocs.length; i++) { 
            if (subprocs[i].findParameterAccess(index)) { 
                return true;
            }
        }
        return false;
    }

    void adjustParameterAccess(int index) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.INPUT) { 
                if ((code[pc+1] & 0xFF) > index) { 
                    code[pc+1] -= 1;
                }
            }
        }
        for (int i = 0; i < subprocs.length; i++) { 
            subprocs[i].adjustParameterAccess(index);
        }
    }

    boolean removeVariable(int i) { 
        if (findVariableAccess(level, i)) { 
            return false;
        }
        int n = vars.length;
        String[] newVars = new String[n-1];
        System.arraycopy(vars, 0, newVars, i, i);
        System.arraycopy(vars, i+1, newVars, i, n-i-1);
        vars = newVars;
        adjustVariableAccess(level, i);
        return true;
    }

    void removeInstruction(int pc) { 
        int len = VirtualMachine.insnSize[code[pc]];
        byte[] newCode = new byte[code.length-len];
        System.arraycopy(code, 0, newCode, 0, pc);
        System.arraycopy(code, pc+len, newCode, pc, newCode.length-pc);
        code = newCode;
        adjustLabels(pc, -len);
    }

    void addInstruction(int pc, int cop) { 
        int len = VirtualMachine.insnSize[cop];
        adjustLabels(pc, len);
        byte[] newCode = new byte[code.length+len];
        System.arraycopy(code, 0, newCode, 0, pc);
        System.arraycopy(code, pc, newCode, pc+len, code.length-pc);
        code = newCode;
        newCode[pc] = (byte)cop;
    }

    void adjustLabels(int pos, int offs) { 
        for (int pc = 0; pc < code.length; pc += VirtualMachine.insnSize[code[pc]]) { 
            if (code[pc] == VirtualMachine.IF || code[pc] == VirtualMachine.GOTO) { 
                int target = Converter.unpackInt(code, pc+1);
                if (target > pos || (target == pos && offs > 0)) { 
                    Converter.packInt(code, pc+1, target+offs);
                }
            }
        }
    }

    int pack(ByteBuffer buf, int offs) {
        byte[] nameBytes = Converter.packString(name);
        buf.extend(offs + 1 + nameBytes.length);
        buf.arr[offs++] = (byte)nameBytes.length;
        System.arraycopy(nameBytes, 0, buf.arr, offs, nameBytes.length);
        offs += nameBytes.length;

        buf.extend(offs + 1);
        buf.arr[offs++] = (byte)vars.length;
        for (int i = 0; i < vars.length; i++) { 
            nameBytes = Converter.packString(vars[i]);
            buf.extend(offs + 1 + nameBytes.length);
            buf.arr[offs++] = (byte)nameBytes.length;
            System.arraycopy(nameBytes, 0, buf.arr, offs, nameBytes.length);
            offs += nameBytes.length;
        }

        buf.extend(offs + 1); 
        buf.arr[offs++] = (byte)level;        
        buf.extend(offs + 4 + code.length);
        Converter.packInt(buf.arr, offs, code.length);
        offs += 4;
        System.arraycopy(code, 0, buf.arr, offs, code.length);
        offs += code.length;

        buf.extend(offs + 1); 
        buf.arr[offs++] = (byte)subprocs.length;        
        for (int i = 0; i < subprocs.length; i++) { 
            offs = subprocs[i].pack(buf, offs);
        }
        return offs;
    }

    int unpack(byte[] record, int offs) { 
        name = Converter.unpackString(record, offs + 1, record[offs] & 0xFF);
        offs += 1 + (record[offs] & 0xFF);
        vars = new String[record[offs++] & 0xFF];
        for (int i = 0; i < vars.length; i++) { 
            vars[i] = Converter.unpackString(record, offs + 1, record[offs] & 0xFF);
            offs += 1 + (record[offs] & 0xFF);
        }
        level = record[offs++] & 0xFF;

        code = new byte[Converter.unpackInt(record, offs)];
        offs += 4;
        System.arraycopy(record, offs, code, 0, code.length);
        offs += code.length;

        subprocs = new Procedure[record[offs++] & 0xFF];
        for (int i = 0; i < subprocs.length; i++) { 
            subprocs[i] = new Procedure(prog, this);
            offs = subprocs[i].unpack(record, offs);
        }
        return offs;
    }
}
            
        





