import java.util.Vector;
import java.util.Random;

public class VirtualMachine { 
    static final int OPERAND_STACK_SIZE = 128;
    static final int CALL_STACK_SIZE    = 128;

    static Float HALF = new Float(1, -1);

    static final int MAX_NUMBER_LEN = 32;
    static final int SIZEOF_FLOAT = 16;
    Float[] opdStack;
    int     oSp;

    Frame[] callStack;
    int     cSp;
    Frame   currFrame;

    static Random  random = new Random();

    Program prog;
    boolean running;

    // load/store
    static final int LITERAL = 0;
    static final int LOAD = 1;
    static final int INPUT = 2;
    static final int STORE = 3;
    static final int DUP = 4;
    static final int SWAP = 5;
    static final int LOAD_E = 6;
    static final int LOAD_PI = 7;

    // operators
    static final int NEG = 8;
    static final int ADD = 9;
    static final int SUB = 10;
    static final int MUL = 11;
    static final int DIV = 12;
    static final int EQ = 13;
    static final int NE = 14;
    static final int GT = 15;
    static final int GE = 16;
    static final int LT = 17;
    static final int LE = 18;

    // functions
    static final int ABS = 19;
    static final int ACOS = 20;
    static final int ASIN = 21;
    static final int ATAN = 22;
    static final int ATAN2 = 23;
    static final int CEIL = 24;
    static final int COS = 25;
    static final int EXP = 26;
    static final int FLOOR = 27;
    static final int FRAC = 28;
    static final int INT = 29;
    static final int LOG = 30;
    static final int LOG10 = 31;
    static final int MAX = 32;
    static final int MIN = 33;
    static final int POW = 34;
    static final int RND = 35;
    static final int ROUND = 36;
    static final int SIN = 37;
    static final int SQRT = 38;
    static final int TAN = 39;

    //statements
    static final int IF = 40;
    static final int CALL = 41;
    static final int RETURN = 42;
    static final int GOTO = 43;

    static final String insnMnemonic[] = {
        "", // LITERAL
        "load",
        "input",
        "store",
        "dup", 
        "swap", 
        "E", 
        "PI", 

        // operators
        "neg", 
        "+",
        "-",
        "*",
        "/", 
        "=", 
        "!=", 
        ">", 
        ">=",
        "<",
        "<=", 

        // functions
        "abs",
        "acos",
        "asin",
        "atan",
        "atan2",
        "ceil",
        "cos",
        "exp",
        "floor",
        "frac",
        "int",
        "log",
        "log10",
        "max",
        "min",
        "pow",
        "rnd",
        "round",
        "sin",
        "sqrt",
        "tan",

        //statements
        "if",
        "call",
        "return",
        "goto"
    };

    static final int insnSize[] = {
        1+SIZEOF_FLOAT,// LITERAL
        3, // "load"
        2, // "input"
        3, // "store"
        1, // "dup"
        1, // "swap"
        1, // "E" 
        1, // "PI"

        // operators
        1, // "neg"
        1, // "+"
        1, // "-"
        1, // "*"
        1, // "/"
        1, // "="
        1, // "!="
        1, // ">"
        1, // ">="
        1, // "<"
        1, // "<="

        // functions
        1, // "abs"
        1, // "acos"
        1, // "asin"
        1, // "atan"
        1, // "atan2"
        1, // "ceil"
        1, // "cos"
        1, // "exp"
        1, // "floor"
        1, // "frac"
        1, // "int"
        1, // "log"
        1, // "log10"
        1, // "max"
        1, // "min"
        1, // "pow"
        1, // "rnd"
        1, // "round"
        1, // "sin"
        1, // "sqrt"
        1, // "tan"

        //statements
        5, // "if"
        3, // "call"
        1, // "return"
        5  // "goto"
    };


    VirtualMachine(Program prog) { 
        this.prog = prog;
        opdStack = new Float[OPERAND_STACK_SIZE];
        callStack = new Frame[CALL_STACK_SIZE];
        callStack[0] = currFrame = new Frame(prog.main, null);
        cSp = 1;
        oSp = 0;
    }

    static String printNumber(Float x) { 
        return x.toString();
    }


    static void packFloat(byte[] buf, int offs, Float x) { 
        Converter.packLong(buf, offs, x.m_Val);
        Converter.packLong(buf, offs+8, x.m_E);
    }

    static Float unpackFloat(byte[] buf, int offs) { 
        return new Float(Converter.unpackLong(buf, offs), Converter.unpackLong(buf, offs+8));
    }

    static Float inputNumber(String s) throws NumberFormatException { 
        return Float.parse(s, 10);
    }

    static void moveLabel(byte[] code, int oldTarget, int newTarget) {        
        for (int pc = 0; pc < code.length; pc += insnSize[code[pc]]) 
        {
            switch (code[pc]) { 
            case IF:
            case GOTO:
                if (Converter.unpackInt(code, pc+1) == oldTarget) { 
                    Converter.packInt(code, pc+1, newTarget);
                }
            }
        }
    }
 
    static int getPcForInstruction(byte[] code, int i) { 
        int pc = 0;
        while (--i >= 0) { 
            pc += insnSize[code[pc]];
        }
        return pc;
    }

    static int getInstructionForPc(byte[] code, int target) { 
        int i = 0;
        for (int pc = 0; pc < target; pc += insnSize[code[pc]]) { 
            i += 1;
        }
        return i;
    }

    static String[] printInstructions(Procedure proc) 
    { 
        byte[] code = proc.code;
        String[] insns = new String[code.length+1];                                   
        int n = 0;
        int[] labels = new int[code.length];
        int nLabels = 0;
        for (int pc = 0; pc < code.length; pc += insnSize[code[pc]]) 
        {
            switch (code[pc]) { 
            case IF:
            case GOTO:
                if (labels[Converter.unpackInt(code, pc+1)] == 0) { 
                    labels[Converter.unpackInt(code, pc+1)] = ++nLabels;
                }
            }
        }
        proc.nLabels = nLabels;
        for (int pc = 0; pc < code.length; pc += insnSize[code[pc]]) 
        { 
            StringBuffer buf = new StringBuffer();
            if (labels[pc] != 0) { 
                buf.append(labels[pc]);
                buf.append(':');
            } else {
                buf.append("  ");
            }
            buf.append(insnMnemonic[code[pc]]);
            switch (code[pc]) { 
            case LOAD:
            case STORE:
            {
                buf.append(' ');
                int level = code[pc+1];
                Procedure scope = proc;
                while (level != scope.level) { 
                    scope = scope.top;
                }
                buf.append(scope.vars[code[pc+2] & 0xFF]);
                break;
            }
            case LITERAL:
                buf.append(printNumber(unpackFloat(code, pc+1)));
                break;
            case INPUT:
                buf.append(' ');
                buf.append(proc.prog.inputParamNames[code[pc+1] & 0xFF]);
                break;
            case CALL:
            {
                buf.append(' ');
                int level = code[pc+1];
                Procedure scope = proc;
                while (level != scope.level) { 
                    scope = scope.top;
                }
                buf.append(scope.subprocs[code[pc+2] & 0xFF].name);
                break;
            }
            case IF:
            case GOTO:
                buf.append(' ');
                buf.append(labels[Converter.unpackInt(proc.code, pc+1)]);
                buf.append(':');
                break;
            }
            insns[n++] = buf.toString();
        }
        insns[n++] = "[* Program End *]";
        String[] result = new String[n];
        System.arraycopy(insns, 0, result, 0, n);
        return result;
    }


    public void run() throws ArithmeticException, StackOverflowException, StackEmptyException { 
        running = true;
        while (cSp != 0 && running) { 
            step();
        }
        running = false;
    }

    final private void push(Float value) throws StackOverflowException { 
        if (oSp == opdStack.length) { 
            throw new StackOverflowException();
        }
        opdStack[oSp++] = value;
    }

    final private Float pop() throws StackEmptyException { 
        if (oSp == 0) { 
            throw new StackEmptyException();
        }
        return opdStack[--oSp];
    }

    public void step() throws ArithmeticException, StackOverflowException, StackEmptyException {         
        byte[] code = currFrame.proc.code;
        Float opd1, opd2;
        switch (code[currFrame.pc]) {
        case LITERAL:
            push(unpackFloat(code, currFrame.pc+1));
            currFrame.pc += VirtualMachine.SIZEOF_FLOAT;
            break;
        case LOAD:
            push(currFrame.display[code[currFrame.pc+1]].vars[code[currFrame.pc+2] & 0xFF]);
            currFrame.pc += 2;            
            break;
        case INPUT:
            push(prog.inputParams[code[currFrame.pc+1] & 0xFF]);
            currFrame.pc += 1;            
            break;            
        case LOAD_E:
            push(Float.E);
            break;            
        case LOAD_PI:
            push(Float.PI);
            break;            
        case STORE:
            currFrame.display[code[currFrame.pc+1]].vars[code[currFrame.pc+2] & 0xFF] = pop();
            currFrame.pc += 2;            
            break;
        case DUP:
            opd1 = pop();
            push(opd1);
            push(opd1);
            break;
        case SWAP:
            opd2 = pop();
            opd1 = pop();
            push(opd2);
            push(opd1);
            break;
        case NEG:
            push(pop().Neg());
            break;                 
        case ADD:
            push(pop().Add(pop()));
            break;
        case SUB:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Sub(opd2));
            break;           
        case MUL:
            push(pop().Mul(pop()));
            break;
        case DIV:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Div(opd2));
            break;            
        case EQ:
            push(pop().Equal(pop()) ? Float.ONE : Float.ZERO);
            break;
        case NE:
            push(pop().Equal(pop()) ? Float.ZERO : Float.ONE);
            break;
        case GT:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Great(opd2) ? Float.ONE : Float.ZERO);
            break;
        case GE:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Great(opd2) || opd1.Equal(opd2) ? Float.ONE : Float.ZERO);
            break;
        case LT:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Less(opd2) ? Float.ONE : Float.ZERO);
            break;
        case LE:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Less(opd2) || opd1.Equal(opd2) ? Float.ONE : Float.ZERO);
            break;
        case ABS:
            push(Float.abs(pop()));
            break;
        case ACOS:
            push(Float.acos(pop()));
            break;
        case ASIN:
            push(Float.asin(pop()));
            break;
        case ATAN:
            push(Float.atan(pop()));
            break;
        case ATAN2:
            opd2 = pop();
            opd1 = pop();
            push(Float.atan2(opd1, opd2));
            break;
        case CEIL:
            push(Float.ceil(pop()));
            break;
        case COS:
            push(Float.cos(pop()));
            break;
        case EXP:
            push(Float.exp(pop()));
            break;
        case FLOOR:
            push(Float.floor(pop()));
            break;
        case FRAC:
            push(Float.Frac(pop()));
            break;
        case INT:
            push(Float.Int(pop()));
            break;
        case LOG:
            push(Float.log(pop()));
            break;
        case LOG10:
            push(Float.log10(pop()));
            break;
        case MAX:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Great(opd2) ? opd1 : opd2);
            break;
        case MIN:
            opd2 = pop();
            opd1 = pop();
            push(opd1.Less(opd2) ? opd1 : opd2);
            break;
        case POW:
            opd2 = pop();
            opd1 = pop();
            push(Float.pow(opd1, opd2));
            break;
        case RND:
            push(new Float(random.nextInt() & Integer.MAX_VALUE).Div(Integer.MAX_VALUE));
            break;
        case ROUND:
            push(Float.floor(pop().Add(HALF)));
            break;           
        case SIN:
            push(Float.sin(pop()));
            break;
        case SQRT:
            push(Float.sqrt(pop()));
            break;
        case TAN:
            push(Float.tan(pop()));
            break;
        case IF:
            if (pop().Equal(Float.ZERO)) { 
                currFrame.pc = Converter.unpackInt(code, currFrame.pc+1);
            } else { 
                currFrame.pc += 5;
            }
            return;
        case CALL:
        {
            if (cSp == callStack.length) { 
                throw new StackOverflowException();
            }
            Procedure proc = currFrame.display[code[currFrame.pc+1]].proc.subprocs[code[currFrame.pc+2] & 0xFF];
            currFrame.pc += 3;
            callStack[cSp++] = currFrame = new Frame(proc, currFrame);
            return;
        }
        case RETURN:
            cSp -= 1;
            if (cSp != 0) {
                currFrame = callStack[cSp-1];
                return;
            }
            break;
        case GOTO:
            currFrame.pc = Converter.unpackInt(code, currFrame.pc+1);
            return;
        }
        currFrame.pc += 1; 
    }
}











